import os
from pathlib import Path

import pytest

from tuxrun.devices import Device
from tuxrun.devices.qemu import QemuArmv5
from tuxrun.devices.fvp import FVPMorelloAndroid
from tuxrun.exceptions import InvalidArgument
from tuxrun.tests import Test


BASE = (Path(__file__) / "..").resolve()


def test_select():
    assert Device.select("qemu-armv5") == QemuArmv5
    assert Device.select("fvp-morello-android") == FVPMorelloAndroid

    with pytest.raises(InvalidArgument):
        Device.select("Hello")


FVP_MORELLO_ANDROID = {
    "tests": [],
    "mcp_fw": "mcp_fw.bin",
    "mcp_romfw": "mcp_romfw.bin",
    "rootfs": "android-nano.img.xz",
    "scp_fw": "scp_fw.bin",
    "scp_romfw": "scp_romfw.bin",
    "uefi": "uefi.bin",
}


@pytest.mark.parametrize(
    "device,args,timeouts,filename",
    [
        (
            "qemu-arm64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-arm64.yaml",
        ),
        (
            "qemu-arm64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {"deploy": 4, "boot": 12},
            "qemu-arm64-timeouts.yaml",
        ),
        (
            "qemu-arm64",
            {
                "command": "",
                "tests": [Test.select("ltp-fcntl-locktests")(timeout=None)],
                "tux_boot_args": "",
                "overlays": [],
                "test_definitions": "testdef.tar.zst",
            },
            {},
            "qemu-arm64-ltp-fcntl-locktests.yaml",
        ),
        (
            "qemu-armv5",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-armv5.yaml",
        ),
        (
            "qemu-armv5",
            {
                "command": "",
                "tests": [Test.select("ltp-fs_bind")(timeout=None)],
                "tux_boot_args": "",
                "overlays": [],
                "test_definitions": "testdef.tar.zst",
            },
            {},
            "qemu-armv5-ltp-fs_bind.yaml",
        ),
        (
            "qemu-armv7",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-armv7.yaml",
        ),
        (
            "qemu-armv7",
            {
                "command": "",
                "tests": [
                    Test.select("ltp-fs_perms_simple")(timeout=4),
                    Test.select("ltp-fsx")(timeout=3),
                    Test.select("ltp-nptl")(timeout=None),
                ],
                "tux_boot_args": "",
                "overlays": [],
                "test_definitions": "testdef.tar.zst",
            },
            {},
            "qemu-armv7-ltp-timeouts.yaml",
        ),
        (
            "qemu-armv7",
            {"tests": [], "tux_boot_args": "", "overlays": [], "kernel": "zImage.xz"},
            {},
            "qemu-armv7-kernel-xz.yaml",
        ),
        (
            "qemu-i386",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-i386.yaml",
        ),
        (
            "qemu-i386",
            {
                "command": "",
                "tests": [Test.select("kunit")(timeout=None)],
                "tux_boot_args": "",
                "overlays": [],
                "test_definitions": "testdef.tar.zst",
            },
            {},
            "qemu-i386-kunit.yaml",
        ),
        (
            "qemu-i386",
            {"tests": [], "tux_boot_args": "", "overlays": [], "kernel": "bzImage.gz"},
            {},
            "qemu-i386-kernel-gz.yaml",
        ),
        (
            "qemu-mips32",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-mips32.yaml",
        ),
        (
            "qemu-mips32",
            {
                "tests": [Test.select("command")(timeout=None)],
                "tux_boot_args": "",
                "overlays": [],
                "command": "cat /proc/cpuinfo",
                "test_definitions": "testdef.tar.zst",
            },
            {},
            "qemu-mips32-command.yaml",
        ),
        (
            "qemu-mips32el",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-mips32el.yaml",
        ),
        (
            "qemu-mips64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-mips64.yaml",
        ),
        (
            "qemu-mips64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-mips64el.yaml",
        ),
        (
            "qemu-ppc32",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-ppc32.yaml",
        ),
        (
            "qemu-ppc64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-ppc64.yaml",
        ),
        (
            "qemu-ppc64le",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-ppc64le.yaml",
        ),
        (
            "qemu-riscv64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-riscv64.yaml",
        ),
        (
            "qemu-sparc64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-sparc64.yaml",
        ),
        (
            "qemu-x86_64",
            {"tests": [], "tux_boot_args": "", "overlays": []},
            {},
            "qemu-x86_64.yaml",
        ),
        (
            "fvp-morello-android",
            FVP_MORELLO_ANDROID,
            {},
            "fvp-morello-android.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("binder")(timeout=None)],
                "parameters": {"USERDATA": "userdata.tar.xz"},
            },
            {},
            "fvp-morello-android-binder.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("bionic")(timeout=None)],
                "parameters": {"USERDATA": "userdata.tar.xz"},
            },
            {},
            "fvp-morello-android-bionic.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("bionic")(timeout=None)],
                "parameters": {
                    "BIONIC_TEST_TYPE": "dynamic",
                    "GTEST_FILTER": "hello",
                    "USERDATA": "userdata.tar.xz",
                },
            },
            {},
            "fvp-morello-android-bionic-params.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("boringssl")(timeout=None)],
                "parameters": {"SYSTEM_URL": "system.tar.xz"},
            },
            {},
            "fvp-morello-android-boringssl.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("boringssl")(timeout=4212)],
                "parameters": {"SYSTEM_URL": "system.tar.xz"},
            },
            {},
            "fvp-morello-android-boringssl-timeouts.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("compartment")(timeout=None)],
                "parameters": {"USERDATA": "userdata.tar.xz"},
            },
            {},
            "fvp-morello-android-compartment.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("device-tree")(timeout=None)],
                "parameters": {},
            },
            {},
            "fvp-morello-android-device-tree.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("dvfs")(timeout=None)],
                "parameters": {},
            },
            {},
            "fvp-morello-android-dvfs.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("libjpeg-turbo")(timeout=None)],
                "parameters": {
                    "SYSTEM_URL": "system.tar.xz",
                    "LIBJPEG_TURBO_URL": "libjpeg.tar.xz",
                },
            },
            {},
            "fvp-morello-android-libjpeg-turbo.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("libpdfium")(timeout=None)],
                "parameters": {
                    "SYSTEM_URL": "system.tar.xz",
                    "PDFIUM_URL": "pdfium.tar.xz",
                },
            },
            {},
            "fvp-morello-android-libpdfium.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("libpng")(timeout=None)],
                "parameters": {"SYSTEM_URL": "system.tar.xz", "PNG_URL": "png.tar.xz"},
            },
            {},
            "fvp-morello-android-libpng.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("lldb")(timeout=None)],
                "parameters": {"LLDB_URL": "lldb.tar.xz", "TC_URL": "toolchain.tar.xz"},
            },
            {},
            "fvp-morello-android-lldb.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("logd")(timeout=None)],
                "parameters": {"USERDATA": "userdata.tar.xz"},
            },
            {},
            "fvp-morello-android-logd.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("multicore")(timeout=None)],
                "parameters": {},
            },
            {},
            "fvp-morello-android-multicore.yaml",
        ),
        (
            "fvp-morello-android",
            {
                **FVP_MORELLO_ANDROID,
                "tests": [Test.select("zlib")(timeout=None)],
                "parameters": {"SYSTEM_URL": "system.tar.xz"},
            },
            {},
            "fvp-morello-android-zlib.yaml",
        ),
        (
            "fvp-morello-busybox",
            {
                "tests": [],
                "mcp_fw": "mcp_fw.bin",
                "mcp_romfw": "mcp_romfw.bin",
                "rootfs": "busybox.img.xz",
                "scp_fw": "scp_fw.bin",
                "scp_romfw": "scp_romfw.bin",
                "uefi": "uefi.bin",
            },
            {},
            "fvp-morello-busybox.yaml",
        ),
        (
            "fvp-morello-oe",
            {
                "tests": [],
                "mcp_fw": "mcp_fw.bin",
                "mcp_romfw": "mcp_romfw.bin",
                "rootfs": "core-image-minimal-morello-fvp.wic",
                "scp_fw": "scp_fw.bin",
                "scp_romfw": "scp_romfw.bin",
                "uefi": "uefi.bin",
            },
            {},
            "fvp-morello-oe.yaml",
        ),
        (
            "fvp-morello-ubuntu",
            {
                "tests": [],
                "mcp_fw": "mcp_fw.bin",
                "mcp_romfw": "mcp_romfw.bin",
                "scp_fw": "scp_fw.bin",
                "scp_romfw": "scp_romfw.bin",
                "uefi": "uefi.bin",
            },
            {},
            "fvp-morello-ubuntu.yaml",
        ),
    ],
)
def test_definition(device, args, timeouts, filename):
    if os.environ.get("TUXRUN_RENDER"):
        obj = Device.select(device)()
        (BASE / "refs" / "definitions" / filename).write_text(
            Device.select(device)().definition(
                device=obj,
                container=None,
                tests_timeout=sum(t.timeout for t in args["tests"]),
                timeouts=timeouts,
                tmpdir=Path("/tmp/tuxrun-ci"),
                **args
            )
        )
    obj = Device.select(device)()
    assert Device.select(device)().definition(
        device=obj,
        container=None,
        tests_timeout=sum(t.timeout for t in args["tests"]),
        timeouts=timeouts,
        tmpdir=Path("/tmp/tuxrun-ci"),
        **args
    ) == (BASE / "refs" / "definitions" / filename).read_text(encoding="utf-8")



# -*- coding: utf-8 -*-
# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import fnmatch

from tuxrun import templates
from tuxrun.exceptions import InvalidArgument


def subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in subclasses(c)]
    )


class Test:
    device: str = ""
    name: str = ""
    timeout: int = 0

    def __init__(self, timeout):
        if timeout:
            self.timeout = timeout

    @classmethod
    def select(cls, name):
        for subclass in subclasses(cls):

            if subclass.name == name:
                return subclass
        raise InvalidArgument(f"Unknown test {name}")

    @classmethod
    def list(cls):
        return sorted([s.name for s in subclasses(cls) if s.name])

    def validate(self, device=device, **kwargs):
        if not fnmatch.fnmatch(device.name, self.device):
            raise InvalidArgument(
                f"Test '{self.name}' not supported on device '{device.name}'"
            )

    def _render(self, filename, **kwargs):
        return templates.tests().get_template(filename).render(**kwargs)


import tuxrun.tests.command  # noqa: E402
import tuxrun.tests.kunit  # noqa: E402
import tuxrun.tests.ltp  # noqa: E402
import tuxrun.tests.morello  # noqa: E402,F401



from tuxrun.results import Results


def test_returns_0_by_default():
    results = Results()
    assert results.ret() == 0


def gen_test(name, result, suite_name="mytestsuite"):
    return f'{{ "lvl": "results", "msg": {{"definition": "{suite_name}", "case": "{name}", "result": "{result}"}}}}'


def test_returns_0_with_no_failures():
    results = Results()
    results.parse(gen_test("test1", "pass"))
    results.parse(gen_test("test2", "pass"))
    assert results.ret() == 0


def test_returns_1_on_failure():
    results = Results()
    results.parse(gen_test("test1", "pass"))
    results.parse(gen_test("test2", "fail"))
    assert results.ret() == 1


def test_returns_invalid_logs():
    results = Results()
    results.parse("{")
    results.parse('{ "lvl": "results", "msg": {"case": "tux", "result": "pass"}}')


def test_data():
    results = Results()
    results.parse(gen_test("test1", "pass"))
    assert results.data["mytestsuite"]["test1"]["result"] == "pass"



# -*- coding: utf-8 -*-
# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

from typing import List, Optional

from tuxrun.exceptions import InvalidArgument
from tuxrun.tests import Test


class MorelloTest(Test):
    template = "morello.yaml.jinja2"
    test_def_name: Optional[str] = None
    parameters: List[str] = []

    def validate(self, device, parameters, **kwargs):
        super().validate(device=device, parameters=parameters, **kwargs)
        missing = set(self.parameters) - set(parameters.keys())
        if missing:
            raise InvalidArgument(f"Missing --parameters {', '.join(missing)}")
        invalid = set(parameters.keys()) - set(self.parameters)
        if invalid:
            raise InvalidArgument(f"Invalid --parameters {', '.join(invalid)}")

    def render(self, **kwargs):
        kwargs["name"] = self.name
        kwargs["timeout"] = self.timeout
        kwargs["test_def_name"] = (
            self.test_def_name if self.test_def_name else self.name
        )

        # remap some parameters
        MAPPINGS = {"LLDB_URL": "LLDB_TESTS_URL", "USERDATA": "USERDATA_URL"}
        for key, value in MAPPINGS.items():
            if key in kwargs["parameters"]:
                kwargs["parameters"][value] = kwargs["parameters"].pop(key)

        return self._render(self.template, **kwargs)


class MorelloAndroidTest(MorelloTest):
    device = "fvp-morello-android"


class MorelloBinder(MorelloAndroidTest):
    name = "binder"
    timeout = 34
    parameters = ["USERDATA"]


class MorelloBionic(MorelloAndroidTest):
    name = "bionic"
    timeout = 1000
    parameters = ["BIONIC_TEST_TYPE", "GTEST_FILTER", "USERDATA"]

    def validate(self, device, parameters, **kwargs):
        if parameters.get("BIONIC_TEST_TYPE") is None:
            parameters["BIONIC_TEST_TYPE"] = "static"
        if parameters["BIONIC_TEST_TYPE"] not in ["dynamic", "static"]:
            raise InvalidArgument("Invalid value for --parameters BIONIC_TEST_TYPE")

        if parameters.get("GTEST_FILTER") is None:
            parameters["GTEST_FILTER"] = (
                "string_nofortify.*-string_nofortify.strlcat_overread:string_nofortify.bcopy:string_nofortify.memmove",
            )

        super().validate(device=device, parameters=parameters, **kwargs)

    def render(self, parameters, **kwargs):
        parameters["TEST_PATHS"] = "nativetest64 nativetestc64"
        parameters["TEST_TYPE"] = parameters.pop(
            "BIONIC_TEST_TYPE",
            "static",
        )
        parameters["GTEST_FILTER"] = parameters.get(
            "GTEST_FILTER",
            "string_nofortify.*-string_nofortify.strlcat_overread:string_nofortify.bcopy:string_nofortify.memmove",
        )

        return super().render(parameters=parameters, **kwargs)


class MorelloBoringSSL(MorelloAndroidTest):
    name = "boringssl"
    timeout = 240
    parameters = ["SYSTEM_URL"]


class MorelloCompartment(MorelloAndroidTest):
    name = "compartment"
    timeout = 15
    parameters = ["USERDATA"]
    test_def_name = "compartment-demo"


class MorelloDeviceTree(MorelloAndroidTest):
    name = "device-tree"
    timeout = 15


class MorelloDvfs(MorelloAndroidTest):
    name = "dvfs"
    timeout = 15


class MorelloFWTS(MorelloTest):
    name = "fwts"
    timeout = 120
    device = "fvp-morello-oe"
    template = "fwts.yaml.jinja2"


class MorelloLibJPEGTurbo(MorelloAndroidTest):
    name = "libjpeg-turbo"
    timeout = 30
    parameters = ["LIBJPEG_TURBO_URL", "SYSTEM_URL"]


class MorelloLibPNG(MorelloAndroidTest):
    name = "libpng"
    timeout = 30
    parameters = ["PNG_URL", "SYSTEM_URL"]


class MorelloLibPDFium(MorelloAndroidTest):
    name = "libpdfium"
    timeout = 30
    parameters = ["PDFIUM_URL", "SYSTEM_URL"]


class MorelloLLDB(MorelloAndroidTest):
    name = "lldb"
    timeout = 30
    parameters = ["LLDB_URL", "TC_URL"]


class MorelloLOGD(MorelloAndroidTest):
    name = "logd"
    timeout = 420
    parameters = ["USERDATA"]


class MorelloMulticore(MorelloAndroidTest):
    name = "multicore"
    timeout = 15
    test_def_name = "multicore-boot"


class Morellozlib(MorelloAndroidTest):
    name = "zlib"
    timeout = 30
    parameters = ["SYSTEM_URL"]



# -*- coding: utf-8 -*-
# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import logging
import yaml

from tuxrun.yaml import yaml_load


LOG = logging.getLogger("tuxrun")


class Results:
    def __init__(self):
        self.__data__ = {}
        self.__ret__ = 0

    def parse(self, line):
        try:
            data = yaml_load(line)
        except yaml.YAMLError:
            LOG.debug(line)
            return
        if not data or not isinstance(data, dict):
            LOG.debug(line)
            return
        if data.get("lvl") != "results":
            return

        test = data.get("msg")
        if not set(["case", "definition"]).issubset(test.keys()):
            LOG.debug(line)
            return

        definition = test.pop("definition")
        case = test.pop("case")
        self.__data__.setdefault(definition, {})[case] = test
        if test["result"] == "fail":
            self.__ret__ = 1

    @property
    def data(self):
        return self.__data__

    def ret(self):
        return self.__ret__



# -*- coding: utf-8 -*-
# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

from tuxrun.tests import Test


class LTPTest(Test):
    device = "qemu-*"
    cmdfile: str = ""

    def render(self, **kwargs):
        kwargs["name"] = self.name
        kwargs["timeout"] = self.timeout
        kwargs["cmdfile"] = (
            self.cmdfile if self.cmdfile else self.name.replace("ltp-", "")
        )

        return self._render("ltp.yaml.jinja2", **kwargs)


class LTPController(LTPTest):
    name = "ltp-controllers"
    timeout = 90


class LTPFcntlLockTests(LTPTest):
    name = "ltp-fcntl-locktests"
    timeout = 2


class LTPFSBind(LTPTest):
    name = "ltp-fs_bind"
    timeout = 15


class LTPFSPermsSimple(LTPTest):
    name = "ltp-fs_perms_simple"
    timeout = 2


class LTPFSX(LTPTest):
    name = "ltp-fsx"
    timeout = 1


class LTPNPTL(LTPTest):
    name = "ltp-nptl"
    timeout = 15


class LTPSmoke(LTPTest):
    name = "ltp-smoke"
    cmdfile = "smoketest"
    timeout = 2
